ANATOMICAL_STRUCTURES = [
    'adrenal_gland_left',
    'adrenal_gland_right',
    'aorta',
    'autochthon_left',
    'autochthon_right',
    'brain',
    'clavicula_left',
    'clavicula_right',
    'colon',
    'duodenum',
    'esophagus',
    'face',
    'femur_left',
    'femur_right',
    'gallbladder',
    'gluteus_maximus_left',
    'gluteus_maximus_right',
    'gluteus_medius_left',
    'gluteus_medius_right',
    'gluteus_minimus_left',
    'gluteus_minimus_right',
    'heart_atrium_left',
    'heart_atrium_right',
    'heart_myocardium',
    'heart_ventricle_left',
    'heart_ventricle_right',
    'hip_left',
    'hip_right',
    'humerus_left',
    'humerus_right',
    'iliac_artery_left',
    'iliac_artery_right',
    'iliac_vena_left',
    'iliac_vena_right',
    'iliopsoas_left',
    'iliopsoas_right',
    'inferior_vena_cava',
    'kidney_left',
    'kidney_right',
    'liver',
    'lung_lower_lobe_left',
    'lung_lower_lobe_right',
    'lung_middle_lobe_right',
    'lung_upper_lobe_left',
    'lung_upper_lobe_right',
    'pancreas',
    'portal_vein_and_splenic_vein',
    'pulmonary_artery',
    'rib_left_1',
    'rib_left_10',
    'rib_left_11',
    'rib_left_12',
    'rib_left_2',
    'rib_left_3',
    'rib_left_4',
    'rib_left_5',
    'rib_left_6',
    'rib_left_7',
    'rib_left_8',
    'rib_left_9',
    'rib_right_1',
    'rib_right_10',
    'rib_right_11',
    'rib_right_12',
    'rib_right_2',
    'rib_right_3',
    'rib_right_4',
    'rib_right_5',
    'rib_right_6',
    'rib_right_7',
    'rib_right_8',
    'rib_right_9',
    'sacrum',
    'scapula_left',
    'scapula_right',
    'small_bowel',
    'spleen',
    'stomach',
    'trachea',
    'urinary_bladder',
    'vertebrae_C1',
    'vertebrae_C2',
    'vertebrae_C3',
    'vertebrae_C4',
    'vertebrae_C5',
    'vertebrae_C6',
    'vertebrae_C7',
    'vertebrae_L1',
    'vertebrae_L2',
    'vertebrae_L3',
    'vertebrae_L4',
    'vertebrae_L5',
    'vertebrae_T1',
    'vertebrae_T10',
    'vertebrae_T11',
    'vertebrae_T12',
    'vertebrae_T2',
    'vertebrae_T3',
    'vertebrae_T4',
    'vertebrae_T5',
    'vertebrae_T6',
    'vertebrae_T7',
    'vertebrae_T8',
    'vertebrae_T9',
]

LABELS = ['age', 'gender', 'institute', 'study_type', 'split']
